/*
              all sorts of post processing of basepair lists.
	      Part of "alifold" and the Vienna RNA package 
                                   
              author: Peter Stadler
*/
/*Last changed Time-stamp: <2016-06-23 11:11:25 ivo> */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <getopt.h>
#include "ali.h"

#define PRIVATE static

extern alidot_options alidot_opt;

void        get_BPstat(pair_info *believe, int BPstat[], int nb);
void        pair_conficts(pair_info *bpairs, int cnt);
int         pair_layers(pair_info *bpairs, int cnt);
pair_info** believable_pairs(pair_info* bpairs, int n, int cnt, int maxL);
stack*      believe2bracket(pair_info *believe, pair_info *bpairs,
			    int nb, int n);
char*       stacks2string(stack *stacks, int n);
void        prune_structure(const char *secstr, char *test);
char**      non_trivial_layers(char **Z, int maxL);
char*       make_structure(char **NT);
pair_info*  make_final_pairs(const pair_info* bpairs, const char *str);
void        dump_pair_file(pair_info *bpairs, int cnt);

PRIVATE int compare_pi_i(const void *pi1, const void *pi2);
PRIVATE int is_isolated_left(const char *test, int n, int i);
PRIVATE int is_isolated_right(const char *test, int n, int i);
PRIVATE int is_legal_substr(const char *test, int l, int r);
PRIVATE int get_left_partner(const char *test, int i);
PRIVATE int get_right_partner(const char *test, int i);

void pair_conficts(pair_info *bpairs, int cnt) {
  int i,j;
  /* for every pair mark whether it conflicts with higher ranked pairs */
  for (i=0; i<cnt; i++) {
    int ii, ij, ji, jj, t;
    bpairs[i].comp=1;
    bpairs[i].rank = i;
    ii = bpairs[i].i; ij = bpairs[i].j;
    if (ij>0) {  /* ii is paired to ij */
      if (ii>ij) {t = ii; ii = ij; ij=t; }
      for (j=0; j<i; j++) {
        if (!bpairs[j].comp) continue;
        /* is (ii.ij) compatible with higher ranked pairs? */
        ji = bpairs[j].i; jj = bpairs[j].j;
        if (ji>jj) {t = ji; ji = jj; jj=t; }
        if ((ii<ji)&&(ji<ij)&&(ij<jj)) {bpairs[i].comp = 0 ; break;}
        if ((ji<ii)&&(ii<jj)&&(jj<ij)) {bpairs[i].comp = 0 ; break;}
        if ((ii==ji)||(ii==jj)||(ij==ji)||(ij==jj))
          {bpairs[i].comp = 0 ; break;}
      }
    } else { /* ii is unpaired */
      for (j=0; ((j<i)&&bpairs[i].comp); j++) 
        if (bpairs[j].comp)
          if ((ii==bpairs[j].i)||(ii==bpairs[j].j))
            bpairs[i].comp = 0;
    }
  }
}

/* For each base pair define a layer to which it belongs as follows:
   (0) The highest-ranking pair belongs to layer 1
   (1) i.j belongs to layer with the minimum L>0 such that it is
       compatible with all higherranking pairs in this layer
   (2) A pair that shares a base with a higher ranking pair belongs to
       layer 0 (junk)
*/

int pair_layers(pair_info *bpairs, int cnt) {
  int i,j,L,maxL;
  maxL=1;
  for (i=0; i<cnt; i++) {
    int ii, ij, ji, jj, t;
    bpairs[i].rank = i;
    ii = bpairs[i].i; ij = bpairs[i].j;
    for(L=1;L<=maxL;L++) {
      bpairs[i].comp=L;
      if (ij>0) {  /* ii is paired to ij */
	if (ii>ij) {t = ii; ii = ij; ij=t; }
	for (j=0; j<i; j++) {
	  if (bpairs[j].comp!=L) continue;
	  /* is (ii.ij) compatible with higher ranked pairs in layer L? */
	  ji = bpairs[j].i; jj = bpairs[j].j;
	  if (ji>jj) {t = ji; ji = jj; jj=t; }
	  if ((ii<ji)&&(ji<ij)&&(ij<jj)) {bpairs[i].comp = 0 ; break;}
	  if ((ji<ii)&&(ii<jj)&&(jj<ij)) {bpairs[i].comp = 0 ; break;}
	  if ((ii==ji)||(ii==jj)||(ij==ji)||(ij==jj))
	    {bpairs[i].comp = 0 ; break;}
	}
      } else { /* ii is unpaired */
	for (j=0; ((j<i)&&(bpairs[i].comp==L)); j++) 
	  if (bpairs[j].comp>0)
	    if ((ii==bpairs[j].i)||(ii==bpairs[j].j))
	      bpairs[i].comp = -1; 
      }
      if ((bpairs[i].comp==L)||(bpairs[i].comp==-1)) break;
      else if (bpairs[i].comp==0) {
	if (L+1>maxL) maxL = L+1;
      }
    }
  }
  for (i=0; i<cnt; i++) if (bpairs[i].comp==-1) bpairs[i].comp=0;
  return maxL;
}

void dump_pair_file(pair_info *bpairs, int cnt) {
  int i,j,l,maxL;
  FILE *dump_file;
  dump_file = fopen("alidot.dump","w");
  if (dump_file) {
    maxL=1;
    for(l=1;l<=maxL;l++) {
      fprintf(dump_file,"###\n### Layer %d\n###\n",l);
      for(i=0;i<cnt;i++) {
	if (bpairs[i].comp>maxL) maxL=bpairs[i].comp;
	if (bpairs[i].comp==l) {
	  fprintf(dump_file,"%5d  %5d %5d  %7.5f %8.4f ",
		  bpairs[i].rank,bpairs[i].i+1,bpairs[i].j+1,
		  bpairs[i].p, bpairs[i].ent);
	  for(j=0;j<8;j++) 
	    fprintf(dump_file,"%3d ",bpairs[i].bp[j]);
	  /* fprintf(dump_file,"%c ", bpairs[i].sym); */
	  fprintf(dump_file,"%4d ",bpairs[i].comp);
	  fprintf(dump_file,"\n");
	}
      }
    }
    fclose(dump_file);
  }
  else fprintf(stderr,"Warning. Could not open dump file.\n");
}

void get_BPstat(pair_info *believe, int BPstat[], int nb) {
  int ncomp,i,j;
  for(i=0;i<nb;i++) {
    for (ncomp=0, j=1; j<=6; j++) if (believe[i].bp[j]) ncomp++;
    BPstat[ncomp]++;
    if (believe[i].bp[0]) BPstat[0]++;
  }
}

pair_info** believable_pairs(pair_info* bpairs, int n, int cnt, int maxL){
  int i, k, nb, L;
  pair_info *believe;
  pair_info **BELL;

  BELL = (pair_info**) space(sizeof(pair_info*)*(maxL+1));
  for(L=1;L<=maxL;L++) {
    /*    printf("L = %d ... ",L); fflush(stdout); */
    for (k=i=0; i< cnt; i++) if (bpairs[i].comp==L) k++;
    believe = (pair_info *) space( (k+1)*sizeof(pair_info));
    for (nb=i=0; i< cnt; i++) {
      if (bpairs[i].comp==L) {
	/* no bps with more than ncmax inconsist */
	if (bpairs[i].bp[0]>alidot_opt.ncmax) continue;  
	believe[nb]    = bpairs[i];
	if (bpairs[i].comp>0) believe[nb].comp=1;
	nb++;
      }
    }
    believe[nb].j = -1;    /* PS_dot_plot needs a 'terminator' entry */
    qsort(believe, nb, sizeof(pair_info), compare_pi_i ); /* sort by i */
    
    /* interrupt stacks at places with too many inconsistencies */
    for (i=0; i<nb; i++) {
      if (believe[i].j<0) continue;
      if (believe[i].bp[0]==0) continue;
      if ((believe[i+1].i==believe[i].i+1) &&
	  (believe[i+1].j==believe[i].j-1)) {  /* consecutive pairs */
	if ((believe[i].bp[0]>1) && (believe[i+1].bp[0]>0))
	  believe[i].comp=2;   /* mark as hole in a stack */
	if ((believe[i].bp[0]>0) && (believe[i+1].bp[0]>1))
	  believe[i+1].comp=2; /* mark as hole in a stack */
      }
    }
    BELL[L] = believe;
    /* printf(" %d pairs\n",nb); */
  }
  return BELL;
}

stack *believe2stacks(pair_info *believe, pair_info *bpairs, int nb, int n) {
  int      i, j, k, ii, ncomp=0, nbad, r, nstack;
  stack    *stacks;
  
  stacks = (stack *) space((n/2)*sizeof(stack));
  nstack=0;
  for (i=0; i<nb; i++) {
    if ((believe[i].j<0)||(believe[i].comp==2)) continue;
    ii=i;
    nbad = believe[i].bp[0];
    r = believe[i].rank;
    for (ncomp=0, j=1; j<=6; j++) if (believe[i].bp[j]) ncomp++;
    while ((believe[ii+1].i==believe[ii].i+1) &&
	   (believe[ii+1].j==believe[ii].j-1)) {
      ii++;
      nbad += believe[ii].bp[0];
      if (believe[ii].rank<r) r=believe[ii].rank;
      for (j=1; j<=6; j++) if (believe[ii].bp[j]) ncomp++; 
    }
    if (ii>i) { /* found a stack */
      stacks[nstack].i = believe[i].i;
      stacks[nstack].j = believe[i].j;
      stacks[nstack].len = ii-i+1;
      stacks[nstack].n_comp = ncomp;
      stacks[nstack].n_bad = nbad;
      stacks[nstack].best = r;
      nstack++;
    }
  }
  
  for (i=ii=0; i<nstack; i++) {
    k = stacks[i].best;
    /* best pair may not have inconsistent seqs */
    if (bpairs[k].bp[0]>0) continue;
    /* special case for stacklength 2 */
    if ((stacks[i].len==2)&&(stacks[i].n_bad>=stacks[i].n_comp)) continue; 
    /* if (stacks[i].len==2) continue; */
    /* minimum prediction probability */
    for (j=1; j<=6; j++) if (bpairs[k].bp[j]) ncomp++; 
    if (bpairs[k].p*ncomp < alidot_opt.pmin ) continue;
    if (i!=ii) stacks[ii] = stacks[i];
    ii++;
  }
  stacks[ii].best=-1;  /* terminator entry */
  
  return stacks;
}

char *stacks2string(stack *stacks, int n) {
  int i,k,nstacks;
  char *cbrack;
  
  cbrack = (char *) space(n+1);
  for (i=0; i<n; i++) cbrack[i]='.';
  nstacks = 0; while(stacks[nstacks].best!=-1) nstacks++; 

  for(i=0;i<nstacks;i++) {   
    for (k=0; k<stacks[i].len; k++) {
      cbrack[stacks[i].i+k]='(';
      cbrack[stacks[i].j-k]=')';
    }
  }
  return cbrack;
}


void prune_structure(const char *secstr, char *test) {
  int i, j, n;
  n = strlen(secstr);
  /* pass 1: remove all pairs from test
     that include positions that are paired in secstr */
  for(i=0;i<n;i++) {
    if ( (test[i]=='(')&&(secstr[i]!='.') ) { 
      j = get_right_partner(test,i);
      test[i]='<';
      test[j]='>';
    }
    else if ( (test[i]==')')&&(secstr[i]!='.') ) { 
      j = get_left_partner(test,i);
      test[i]='>';
      test[j]='<';
    }
  }
  /* pass 2: remove all remaining isolated base pairs */
  
  for(i=0;i<n;i++) {
    if ( is_isolated_left(test,n,i) ) {
      j = get_right_partner(test,i);
      if (j<0) exit(101);
      test[i]='<';
      test[j]='>';
    }
    if ( is_isolated_right(test,n,i) ) {
      j = get_left_partner(test,i);
      if (j<0) exit(101);
      test[i]='>';
      test[j]='<';
    }
  }
}

char **non_trivial_layers(char **Z, int maxL) {
  int n,i,j,k,l,b,nnT;
  char **NT;
  
  n = strlen(Z[1]);

  NT = (char **) space(sizeof(char*)*(maxL+2));
  NT[1] = strdup(Z[1]);

  k=1;
  for(i=2;i<=maxL;i++) {
    b=0; for(j=0;j<n;j++) if (Z[i][j]=='(') b++;
    if (b) {
      k++;
      NT[k] = strdup(Z[i]);
      for(j=0;j<n;j++) {
	if ( (NT[k][j]!='(')&&(NT[k][j]!=')')) NT[k][j]='.';
      }
    }
  }
  nnT=k;
  
  /* now try to move base pairs up as far as possible */
  for(k=2;k<=nnT;k++) {
    for(i=0;i<n;i++) {
      if (NT[k][i]=='(') {
	j = get_right_partner(NT[k],i);
	for(l=1;l<k;l++) {
	  if ( is_legal_substr( NT[l],i,j )) {
	    if ( (NT[l][i]=='.')&&(NT[l][j]=='.') ) {
	      NT[l][i]=NT[k][i]; NT[k][i]='.';
	      NT[l][j]=NT[k][j]; NT[k][j]='.';
	    }
	    else {
	      fprintf(stderr,"Fatal error. No idea how this could happen ;-(");
	      exit(102);
	    }
	  }
	}    
      }
    }
  }

  /* next we remove all empty layers */

  for(k=1,l=0;k<=nnT;k++) {
    for(j=0,b=0;j<n;j++) { if (NT[k][j]!='.') b=1; }
    if (b) {
      l++;
      if (k!=l) { NT[l] = NT[k];  NT[k]=NULL; } 
    }
    else {
      free(NT[k]); NT[k]=NULL;
    }
  }
  
  return NT;
}

char *make_structure(char **NT) {
  int j,l,n,nl;
  char *str;
  char R,L;
  nl=1;while(NT[nl]!=NULL) nl++; 
  str = strdup(NT[1]);
  n = strlen(str);
  for(l=2;l<nl;l++) {
    switch(l) {
    case 2:  L = '[';  R = ']' ;  break;
    case 3:  L = '<';  R = '>' ;  break;
    case 4:  L = '{';  R = '}' ;  break;
    default: L = '|';  R = '|' ;
    }
    for(j=0;j<n;j++) {
      if ( NT[l][j]=='(' ) str[j] = L;
      if ( NT[l][j]==')' ) str[j] = R;
    }
  }
  return str;
}

extern pair_info*  make_final_pairs(const pair_info* bpairs, const char *str) {
  pair_info* final;
  int i,j,b,np;
  char R,L;

  np=0; while(bpairs[np].j>0) np++;

  final = (pair_info *) space(sizeof(pair_info)*(np+2));

  for(i=0,j=0;i<np;i++) {
    int ii, jj, tmp;
    ii = bpairs[i].i; jj = bpairs[i].j;
    if (jj<ii) { tmp=ii; ii=jj; jj=tmp; }
    L=str[ii]; R=str[jj];
    b=0;
    if     ( (L=='(')&&(R==')') ) b=1;
    else if ( (L=='[')&&(R==']') ) b=1;
    else if ( (L=='<')&&(R=='>') ) b=1;
    else if ( (L=='{')&&(R=='}') ) b=1;
    if (b) {
      final[j]=bpairs[i];
      j++;
    }
  }
  final[j].j=-1; /* terminator entry */
  
  return final;
}
 

/*------------------------------------------------------------------------*/

PRIVATE int get_left_partner(const char *test, int i) {
  int k,j;
  if (test[i]!=')') return -1;
  k=1;j=i;
  while(k!=0) {
    j--;
    if (j<0) return -1;
    if (test[j]==')') k++;
    if (test[j]=='(') k--;
  }
  return j;
}

PRIVATE int get_right_partner(const char *test, int i) {
  int k,j;
  if (test[i]!='(') return -1;
  k=1;j=i;
  while(k!=0) {
    j++;
    if (test[j]=='(') k++;
    if (test[j]==')') k--;
    if (test[j]=='\0') return -1;
  }
  return j;
}

PRIVATE int is_legal_substr(const char *test, int l, int r) {
  /* test if test[l+1 .. r-1] is a legal secondary structure */
  int i,k;
  k=0;
  for (i=l+1; i<=r-1; i++) {
    if (test[i]=='(') k++;
    if (test[i]==')') k--;
    if (k<0) return 0;
  }
  if (k!=0) return 0;
  else return 1;
}
  
PRIVATE int is_isolated_left(const char *test, int n, int i) {
  if (test[i]!='(') return 0;
  if (test[i+1]=='(') return 0;
  if (i>0) 
    if (test[i-1]=='(') return 0;
  return 1;
}

PRIVATE int is_isolated_right(const char *test, int n, int i) {
  if (test[i]!=')') return 0;
  if (test[i-1]==')') return 0;
  if (i<n-1)
    if (test[i+1]==')') return 0;
  return 1;
}
    
PRIVATE int compare_pi_i(const void *pi1, const void *pi2) { /* for sort by i */
  return(((pair_info *) pi1)->i - ((pair_info *) pi2)->i);
}

/*------------------------------------------------------------------------*/


/* End of file */
